@@ -1,4 +1,6 @@
---- IMAP functions.
+---
+-- IMAP functions.
+--
 -- @copyright Same as Nmap--See http://nmap.org/book/man-legal.html
 
 module(... or "imap", package.seeall)
@@ -19,7 +21,7 @@ function capabilities(host, port)
    local capas = {}
    socket:set_timeout(10000)
    local proto = (port.version and port.version.service_tunnel == "ssl" and "ssl") or "tcp"
-   if not socket:connect(host.ip, port.number, proto) then return nil, "Could Not Connect" end
+   if not socket:connect(host, port, proto) then return nil, "Could Not Connect" end
 
    local status, line = socket:receive_lines(1)
    if not string.match(line, "^[%*] OK") then return nil, "No Response" end
